@@ -1,34 +1,41 @@
 // Include Files							/*{{{*/
 #include <config.h>
 
-#include <apt-pkg/acquire.h>
 #include <apt-pkg/acquire-item.h>
+#include <apt-pkg/acquire.h>
 #include <apt-pkg/cacheset.h>
-#include <apt-pkg/cmndline.h>
 #include <apt-pkg/clean.h>
+#include <apt-pkg/cmndline.h>
 #include <apt-pkg/configuration.h>
 #include <apt-pkg/error.h>
 #include <apt-pkg/fileutl.h>
 #include <apt-pkg/strutl.h>
 
+#include <apt-private/acqprogress.h>
 #include <apt-private/private-cachefile.h>
 #include <apt-private/private-download.h>
 #include <apt-private/private-output.h>
 #include <apt-private/private-utils.h>
-#include <apt-private/acqprogress.h>
 
 #include <fstream>
 #include <string>
 #include <vector>
 
-#include <unistd.h>
-#include <sys/types.h>
-#include <pwd.h>
 #include <fcntl.h>
+#include <pwd.h>
+#include <sys/types.h>
+#include <unistd.h>
+#ifdef HAVE_VFS_H
 #include <sys/vfs.h>
-#include <sys/statvfs.h>
-#include <sys/stat.h>
+#else
+#ifdef HAVE_PARAMS_H
+#include <sys/params.h>
+#endif
+#include <sys/mount.h>
+#endif
 #include <errno.h>
+#include <sys/stat.h>
+#include <sys/statvfs.h>
 
 #include <apti18n.h>
 									/*}}}*/
@@ -66,14 +73,12 @@ bool AuthPrompt(std::vector<std::string> const &UntrustedList, bool const Prompt
    if (_config->FindI("quiet",0) < 2
        && _config->FindB("APT::Get::Assume-Yes",false) == false)
    {
-      c2out << _("Install these packages without verification?") << std::flush;
-      if (!YnPrompt(false))
+      if (!YnPrompt(_("Install these packages without verification?"), false))
          return _error->Error(_("Some packages could not be authenticated"));
 
       return true;
    }
    else if (_config->FindB("APT::Get::Force-Yes",false) == true) {
-      _error->Warning(_("--force-yes is deprecated, use one of the options starting with --allow instead."));
       return true;
    }
 
@@ -109,7 +114,7 @@ bool AcquireRun(pkgAcquire &Fetcher, int const PulseInterval, bool * const Failu
       uri.User.clear();
       uri.Password.clear();
       std::string descUri = std::string(uri);
-      _error->Error(_("Failed to fetch %s  %s\n"), descUri.c_str(),
+      _error->Error(_("Failed to fetch %s  %s"), descUri.c_str(),
 	    (*I)->ErrorText.c_str());
 
       if (Failure != NULL)
@@ -144,7 +149,7 @@ bool CheckFreeSpaceBeforeDownload(std::string const &Dir, unsigned long long Fet
       {
 	 struct statfs Stat;
 	 if (statfs(Dir.c_str(),&Stat) != 0
-#if HAVE_STRUCT_STATFS_F_TYPE
+#ifdef HAVE_STRUCT_STATFS_F_TYPE
 	       || Stat.f_type != RAMFS_MAGIC
 #endif
 	    )
@@ -206,6 +211,7 @@ bool DoDownload(CommandLine &CmdL)
 	       I->Owner->FileSize << ' ' << I->Owner->HashSum() << std::endl;
       return true;
    }
+   auto const storecopy = storefile;
 
    if (_error->PendingError() == true || CheckAuth(Fetcher, false) == false)
       return false;
@@ -215,18 +221,22 @@ bool DoDownload(CommandLine &CmdL)
       return false;
 
    // copy files in local sources to the current directory
+   i = 0;
    for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin(); I != Fetcher.ItemsEnd(); ++I)
    {
-      std::string const filename = cwd + flNotDir((*I)->DestFile);
+      if (dynamic_cast<pkgAcqArchive*>(*I) == nullptr)
+	 continue;
+
       if ((*I)->Local == true &&
-          filename != (*I)->DestFile &&
-          (*I)->Status == pkgAcquire::Item::StatDone)
+	  (*I)->Status == pkgAcquire::Item::StatDone &&
+	  (*I)->DestFile != storecopy[i])
       {
 	 std::ifstream src((*I)->DestFile.c_str(), std::ios::binary);
-	 std::ofstream dst(filename.c_str(), std::ios::binary);
+	 std::ofstream dst(storecopy[i].c_str(), std::ios::binary);
 	 dst << src.rdbuf();
-	 chmod(filename.c_str(), 0644);
+	 chmod(storecopy[i].c_str(), 0644);
       }
+      ++i;
    }
    return Failed == false;
 }
@@ -242,7 +252,7 @@ bool DoChangelog(CommandLine &CmdL)
    APT::VersionList verset = APT::VersionList::FromCommandLine(Cache,
 		CmdL.FileList + 1, APT::CacheSetHelper::CANDIDATE, helper);
    if (verset.empty() == true)
-      return false;
+      return _error->Error(_("No packages found"));
 
    bool const downOnly = _config->FindB("APT::Get::Download-Only", false);
    bool const printOnly = _config->FindB("APT::Get::Print-URIs", false);
@@ -336,12 +346,12 @@ bool DoClean(CommandLine &)
  class LogCleaner : public pkgArchiveCleaner
 {
    protected:
-      virtual void Erase(const char *File, std::string Pkg, std::string Ver,struct stat &St) APT_OVERRIDE
+      virtual void Erase(int const dirfd, char const * const File, std::string const &Pkg, std::string const &Ver,struct stat const &St) APT_OVERRIDE
       {
 	 c1out << "Del " << Pkg << " " << Ver << " [" << SizeToStr(St.st_size) << "B]" << std::endl;
 
 	 if (_config->FindB("APT::Get::Simulate") == false)
-	    RemoveFile("Cleaner::Erase", File);
+	    RemoveFileAt("Cleaner::Erase", dirfd, File);
       };
 };
 bool DoAutoClean(CommandLine &)
